/*
 * Copyright (C) 2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file select.c The select mode of adu. */

#include <dirent.h> /* readdir() */
#include "adu.h"
#include "gcc-compat.h"
#include "cmdline.h"
#include "fd.h"
#include "string.h"
#include "error.h"
#include "portable_io.h"

/** Global dir count. */
static uint64_t num_dirs;
/** Global files count. */
static uint64_t num_files;
/** Global bytes count. */
static uint64_t num_bytes;

/** The decimal representation of an uint64_t never exceeds that size. */
#define FORMATED_VALUE_SIZE 25

/* these get filled in by the select command. */
static char count_unit_buf[4] = "( )", size_unit_buf[4] = "( )";

enum global_stats_flags {
	GSF_PRINT_DIRNAME = 1,
	GSF_PRINT_BYTES = 2,
	GSF_PRINT_FILES = 4,
	GSF_COMPUTE_SUMMARY = 8,
};

struct global_stats_info {
	uint32_t count;
	int ret;
	int osl_errno;
	enum global_stats_flags flags;
};

enum user_stats_flags {
	USF_PRINT_DIRNAME = 1,
	USF_PRINT_BYTES = 2,
	USF_PRINT_FILES = 4,
	USF_COMPUTE_SUMMARY = 8,
};

struct user_stats_info {
	uint32_t count;
	enum user_stats_flags flags;
	int ret;
	int osl_errno;
	struct user_info *ui;
};

static const uint64_t size_unit_divisors[] = {
	[size_unit_arg_b] = 1ULL,
	[size_unit_arg_k] = 1024ULL,
	[size_unit_arg_m] = 1024ULL * 1024ULL,
	[size_unit_arg_g] = 1024ULL * 1024ULL * 1024ULL,
	[size_unit_arg_t] = 1024ULL * 1024ULL * 1024ULL * 1024ULL,
};

static const uint64_t count_unit_divisors[] = {

	[count_unit_arg_n] = 1ULL,
	[count_unit_arg_k] = 1000ULL,
	[count_unit_arg_m] = 1000ULL * 1000ULL,
	[count_unit_arg_g] = 1000ULL * 1000ULL * 1000ULL,
	[count_unit_arg_t] = 1000ULL * 1000ULL * 1000ULL * 1000ULL,
};

static const char size_unit_abbrevs[] = " BKMGT";
static const char count_unit_abbrevs[] = "  kmgt";

static enum enum_size_unit format_size_value(enum enum_size_unit unit,
		uint64_t value, int print_unit, char *result)
{
	enum enum_size_unit u = unit;
	char unit_buf[2] = "\0\0";

	if (unit == size_unit_arg_h) /* human readable */
		for (u = size_unit_arg_b; u < size_unit_arg_t &&
				value > size_unit_divisors[u + 1]; u++)
			; /* nothing */
	if (print_unit)
		unit_buf[0] = size_unit_abbrevs[u];
	sprintf(result, "%llu%s",
		(long long unsigned)value / size_unit_divisors[u], unit_buf);
	return u;
}

static enum enum_count_unit format_count_value(enum enum_count_unit unit,
		uint64_t value, int print_unit, char *result)
{
	enum enum_count_unit u = unit;
	char unit_buf[2] = "\0\0";

	if (unit == count_unit_arg_h) /* human readable */
		for (u = count_unit_arg_n; u < count_unit_arg_t &&
				value > count_unit_divisors[u + 1]; u++)
			; /* nothing */
	if (print_unit)
		unit_buf[0] = count_unit_abbrevs[u];
	sprintf(result, "%llu%s",
		(long long unsigned)value / count_unit_divisors[u], unit_buf);
	return u;
}

static int get_dir_name_by_number(uint64_t *dirnum, char **name)
{
	char *result = NULL, *tmp;
	struct osl_row *row;
	uint64_t val = *dirnum;
	struct osl_object obj = {.data = &val, .size = sizeof(val)};
	int ret;

again:
	ret = osl(osl_get_row(dir_table, DT_NUM, &obj, &row));
	if (ret < 0)
		goto out;
	ret = osl(osl_get_object(dir_table, row, DT_NAME, &obj));
	if (ret < 0)
		goto out;
	if (result) {
		tmp = make_message("%s/%s", (char *)obj.data, result);
		free(result);
		result = tmp;
	} else
		result = adu_strdup((char *)obj.data);
	ret = osl(osl_get_object(dir_table, row, DT_PARENT_NUM, &obj));
	if (ret < 0)
		goto out;
	val = *(uint64_t *)obj.data;
	if (val)
		goto again;
out:
	if (ret < 0) {
		free(result);
		*name = NULL;
	} else
		*name = result;
	return ret;
}

static int get_dir_name_of_row(struct osl_row *dir_table_row, char **name)
{
	struct osl_object obj;
	int ret;
	char *this_dir, *prefix = NULL;

	*name = NULL;
	ret = osl(osl_get_object(dir_table, dir_table_row, DT_NAME, &obj));
	if (ret < 0)
		return ret;
	this_dir = adu_strdup((char *)obj.data);
	ret = osl(osl_get_object(dir_table, dir_table_row, DT_PARENT_NUM, &obj));
	if (ret < 0)
		goto out;
	if (!*(uint64_t *)obj.data) {
		*name = this_dir;
		return 1;
	}
	ret = get_dir_name_by_number((uint64_t *)obj.data, &prefix);
	if (ret < 0)
		goto out;
	*name = make_message("%s/%s", prefix, this_dir);
	free(prefix);
	ret = 1;
out:
	free(this_dir);
	return ret;
}
static int user_stats_loop_function(struct osl_row *row, void *data)
{
	struct user_stats_info *usi = data;
	struct osl_object obj;
	int ret, summary = usi->flags & GSF_COMPUTE_SUMMARY;
	char formated_value[FORMATED_VALUE_SIZE];

	check_signals();
	if (!usi->count && !summary) {
		ret = -E_LOOP_COMPLETE;
		goto err;
	}
	if (summary || (usi->count && (usi->flags & USF_PRINT_FILES))) {
		uint64_t files;
		ret = osl(osl_get_object(usi->ui->table, row, UT_FILES, &obj));
		if (ret < 0)
			goto err;
		files = *(uint64_t *)obj.data;
		if (usi->count && (usi->flags & USF_PRINT_FILES)) {
			format_count_value(conf.count_unit_arg, files,
				conf.count_unit_arg == count_unit_arg_h,
				formated_value);
			printf("\t%s%s", formated_value,
				(usi->flags & (USF_PRINT_BYTES | USF_PRINT_DIRNAME))?
					"\t" : "\n"
			);
		}
		if (summary)
			usi->ui->files += files;
	}
	if (summary || (usi->count && (usi->flags & USF_PRINT_BYTES))) {
		uint64_t bytes;
		ret = osl(osl_get_object(usi->ui->table, row, UT_BYTES, &obj));
		if (ret < 0)
			goto err;
		bytes = *(uint64_t *)obj.data;
		if (usi->count && (usi->flags & USF_PRINT_BYTES)) {
			format_size_value(conf.size_unit_arg, bytes,
				conf.size_unit_arg == size_unit_arg_h,
				formated_value);
			printf("%s%s%s",
				(usi->flags & USF_PRINT_FILES)? "" : "\t",
				formated_value,
				usi->flags & USF_PRINT_DIRNAME?  "\t" : "\n"
			);
		}
		if (summary) {
			usi->ui->bytes += bytes;
			usi->ui->dirs++;
		}

	}
	if (usi->count && (usi->flags & USF_PRINT_DIRNAME)) {
		char *dirname;
		ret = osl(osl_get_object(usi->ui->table, row, UT_DIR_NUM, &obj));
		if (ret < 0)
			goto err;
		ret = get_dir_name_by_number((uint64_t *)obj.data, &dirname);
		if (ret < 0)
			goto err;
		printf("%s%s\n",
			(usi->flags & (USF_PRINT_BYTES | USF_PRINT_FILES))? "" : "\t",
			dirname);
		free(dirname);
	}
	if (usi->count > 0)
		usi->count--;
	return 1;
err:
	usi->ret = ret;
	usi->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return -1;
}

static int global_stats_loop_function(struct osl_row *row, void *data)
{
	struct global_stats_info *gsi = data;
	struct osl_object obj;
	char *dirname, formated_value[FORMATED_VALUE_SIZE];
	int ret, summary = gsi->flags & GSF_COMPUTE_SUMMARY;

	check_signals();
	if (!gsi->count && !summary) {
		ret = -E_LOOP_COMPLETE;
		goto err;
	}
	if (summary || (gsi->count && (gsi->flags & GSF_PRINT_FILES))) {
		uint64_t files;
		ret = osl(osl_get_object(dir_table, row, DT_FILES, &obj));
		if (ret < 0)
			goto err;
		files = *(uint64_t *)obj.data;
		if (gsi->count && (gsi->flags & GSF_PRINT_FILES)) {
			format_count_value(conf.count_unit_arg, files,
				conf.count_unit_arg == count_unit_arg_h,
				formated_value);
			printf("\t%s%s", formated_value,
				(gsi->flags & (GSF_PRINT_BYTES | GSF_PRINT_DIRNAME))?
				"\t" : "\n");
		}
		if (summary)
			num_files += files;
	}
	if (summary || (gsi->count && (gsi->flags & GSF_PRINT_BYTES))) {
		uint64_t bytes;
		ret = osl(osl_get_object(dir_table, row, DT_BYTES, &obj));
		if (ret < 0)
			goto err;
		bytes = *(uint64_t *)obj.data;
		if (gsi->count && (gsi->flags & GSF_PRINT_BYTES)) {
			format_size_value(conf.size_unit_arg, bytes,
				conf.size_unit_arg == size_unit_arg_h,
				formated_value);
			printf("%s%s%s",
				(gsi->flags & GSF_PRINT_FILES)? "" : "\t",
				formated_value,
				(gsi->flags & GSF_PRINT_DIRNAME)? "\t" : "\n"
			);
		}
		if (summary) {
			num_bytes += bytes;
			num_dirs++;
		}
	}
	if (gsi->count && (gsi->flags & GSF_PRINT_DIRNAME)) {
		ret = get_dir_name_of_row(row, &dirname);
		if (ret < 0)
			goto err;
		printf("%s%s\n",
			(gsi->flags & (GSF_PRINT_BYTES | GSF_PRINT_FILES))? "" : "\t",
			dirname);
		free(dirname);
	}
	if (gsi->count > 0)
		gsi->count--;
	return 1;
err:
	gsi->ret = ret;
	gsi->osl_errno = (ret == -E_OSL)? osl_errno : 0;
	return -1;
}

static int check_loop_return(int ret, int loop_ret, int loop_osl_errno)
{
	if (ret >= 0)
		return ret;
	assert(ret == -E_OSL);
	if (osl_errno != E_OSL_LOOP)
		/* error not caused by loop function returning negative. */
		return ret;
	assert(loop_ret < 0);
	if (loop_ret == -E_LOOP_COMPLETE) /* no error */
		return 1;
	if (loop_ret == -E_OSL) { /* osl error in loop function */
		assert(loop_osl_errno);
		osl_errno = loop_osl_errno;
	}
	return loop_ret;
}

static int adu_loop_reverse(struct osl_table *t, unsigned col_num, void *private_data,
		osl_rbtree_loop_func *func, int *loop_ret, int *loop_osl_errno)
{
	int ret = osl(osl_rbtree_loop_reverse(t, col_num, private_data, func));
	return check_loop_return(ret, *loop_ret, *loop_osl_errno);
}

static void print_global_summary(void)
{
	char d[FORMATED_VALUE_SIZE], f[FORMATED_VALUE_SIZE],
		s[FORMATED_VALUE_SIZE];
	enum enum_count_unit ud, uf;
	enum enum_size_unit us;

	ud = format_count_value(conf.count_unit_arg, num_dirs, 0, d);
	uf = format_count_value(conf.count_unit_arg, num_files, 0, f);
	us = format_size_value(conf.size_unit_arg, num_bytes, 0, s);

	printf("Global summary "
		"(dirs(%c)/files(%c)/size(%c))\n"
		"\t%s\t%s\t%s\n\n",
		count_unit_abbrevs[ud],
		count_unit_abbrevs[uf],
		size_unit_abbrevs[us],
		d, f, s
	);

}

static int print_user_summary_line(struct user_info *ui, __a_unused void *data)
{
	char formated_dir_count[FORMATED_VALUE_SIZE],
		formated_file_count[FORMATED_VALUE_SIZE],
		formated_bytes[FORMATED_VALUE_SIZE ];

	format_count_value(conf.count_unit_arg, ui->dirs,
		conf.count_unit_arg == count_unit_arg_h,
		formated_dir_count);
	format_count_value(conf.count_unit_arg, ui->files,
		conf.count_unit_arg == count_unit_arg_h,
		formated_file_count);
	format_size_value(conf.size_unit_arg, ui->bytes,
		conf.size_unit_arg == size_unit_arg_h,
		formated_bytes);
	printf("\t%s\t%u\t%s\t%s\t%s\n",
		ui->pw_name? ui->pw_name : "?",
		(unsigned)ui->uid,
		formated_dir_count,
		formated_file_count,
		formated_bytes
	);
	return 1;
}

static void print_user_summary(void)
{
	printf("User summary "
		"(pw_name/uid/dirs%s/files%s/size%s):\n",
		count_unit_buf, count_unit_buf, size_unit_buf);
	for_each_admissible_user(print_user_summary_line, NULL);
}

static int print_user_stat(struct user_info *ui, __a_unused void *data)
{
	int ret;
	struct user_stats_info usi = {
		.count = conf.limit_arg,
		.ui = ui
	};

	usi.flags = USF_PRINT_DIRNAME | USF_PRINT_BYTES | USF_COMPUTE_SUMMARY;
	printf("%s (uid %u), by size%s:\n",
		ui->pw_name? ui->pw_name : "?", (unsigned)ui->uid,
		size_unit_buf);
	ret = adu_loop_reverse(ui->table, UT_BYTES, &usi, user_stats_loop_function,
		&usi.ret, &usi.osl_errno);
	if (ret < 0)
		return ret;
	printf("\n%s (uid %u), by file count%s:\n",
		ui->pw_name? ui->pw_name : "?", (unsigned)ui->uid,
		count_unit_buf);
	usi.count = conf.limit_arg,
	usi.flags = USF_PRINT_DIRNAME | USF_PRINT_FILES;
	ret = adu_loop_reverse(ui->table, UT_FILES, &usi, user_stats_loop_function,
		&usi.ret, &usi.osl_errno);
	if (ret < 0)
		return ret;
	printf("\n");
	return 1;
}

static int print_user_stats(void)
{
	return for_each_admissible_user(print_user_stat, NULL);
}

static int print_statistics(void)
{
	int ret;
	struct global_stats_info gsi = {
		.count = conf.limit_arg,
		.flags = GSF_PRINT_DIRNAME | GSF_PRINT_BYTES | GSF_COMPUTE_SUMMARY
	};

	printf("By size%s:\n",
		size_unit_buf);
	ret = adu_loop_reverse(dir_table, DT_BYTES, &gsi,
		global_stats_loop_function, &gsi.ret, &gsi.osl_errno);
	if (ret < 0)
		return ret;
	printf("\n");

	gsi.count = conf.limit_arg;
	gsi.flags = GSF_PRINT_DIRNAME | GSF_PRINT_FILES;
	printf("By file count%s:\n",
		count_unit_buf);
	ret = adu_loop_reverse(dir_table, DT_FILES, &gsi,
		global_stats_loop_function, &gsi.ret, &gsi.osl_errno);
	if (ret < 0)
		return ret;
	printf("\n");
	print_global_summary();
	print_user_stats();
	print_user_summary();
	return 1;
}

static int read_uid_file(void)
{
	size_t size;
	uint32_t n;
	char *filename = get_uid_list_name(), *map;
	int ret = mmap_full_file(filename, O_RDONLY, (void **)&map, &size, NULL);
	unsigned bits;

	if (ret < 0) {
		INFO_LOG("failed to map %s\n", filename);
		free(filename);
		return ret;
	}
	num_uids = size / 4;
	INFO_LOG("found %u uids in %s\n", (unsigned)num_uids, filename);
	free(filename);
	/*
	 * Compute number of hash table bits. The hash table size must be a
	 * power of two and larger than the number of uids.
	 */
	bits = 2;
	while (1 << bits < num_uids)
		bits++;
	create_hash_table(bits);
	for (n = 0; n < num_uids; n++) {
		uint32_t uid = read_u32(map + n * sizeof(uid));
		ret = search_uid(uid, OPEN_USER_TABLE, NULL);
		if (ret < 0)
			goto out;
	}
out:
	adu_munmap(map, size);
	return ret;
}

int com_select(void)
{
	int ret;

	if (conf.count_unit_arg != count_unit_arg_h)
		count_unit_buf[1] = count_unit_abbrevs[conf.count_unit_arg];
	else
		count_unit_buf[0] = '\0';
	if (conf.size_unit_arg != size_unit_arg_h)
		size_unit_buf[1] = size_unit_abbrevs[conf.size_unit_arg];
	else
		size_unit_buf[0] = '\0';

	ret = open_dir_table(0);
	if (ret < 0)
		return ret;
	check_signals();
	ret = read_uid_file();
	if (ret < 0)
		return ret;
	check_signals();
	ret = print_statistics();
	close_all_tables();
	return ret;
}
